@@ -4,15 +4,18 @@
 #' \code{rrapply} is a reimplemented and extended version of \code{\link{rapply}} to recursively apply a function \code{f} to 
 #' a set of elements of a list and deciding \emph{how} the result is structured. 
 #' 
-#' @section List pruning:
+#' @section How to structure result:
 #' In addition to \code{\link{rapply}}'s modes to set \code{how} equal to \code{"replace"}, \code{"list"} or \code{"unlist"}, 
-#' three choices \code{"prune"}, \code{"flatten"} and \code{"melt"} are available. \code{how = "prune"} filters all list elements not subject to 
-#' application of \code{f} from the list \code{object}. The original list structure is retained, similar to the non-pruned options 
-#' \code{how = "replace"} or \code{how = "list"}. \code{how = "flatten"} is an efficient way to return a flattened unnested version 
-#' of the pruned list. \code{how = "melt"} returns a melted data.frame of the pruned list, each row contains the path of a single terminal node in 
-#' the pruned list at depth layers \code{L1}, \code{L2}, and so on. The list-column \code{"value"} contains the values at the terminal nodes and is 
-#' equivalent to the flattened list returned by \code{how = "flatten"}. If no list names are present, the node names in the data.frame default to
-#' the indices of the list elements \code{"..1"}, \code{"..2"}, etc. 
+#' four choices \code{"prune"}, \code{"flatten"}, \code{"melt"} and \code{"unmelt"} are available. \code{how = "prune"} filters 
+#' all list elements not subject to application of \code{f} from the list \code{object}. The original list structure is retained, 
+#' similar to the non-pruned options \code{how = "replace"} or \code{how = "list"}. \code{how = "flatten"} is an efficient way to 
+#' return a flattened unnested version of the pruned list. \code{how = "melt"} returns a melted data.frame of the pruned list, 
+#' each row contains the path of a single terminal node in the pruned list at depth layers \code{L1}, \code{L2}, and so on. The list-column 
+#' \code{"value"} contains the values at the terminal nodes and is equivalent to the flattened list returned by \code{how = "flatten"}. 
+#' If no list names are present, the node names in the data.frame default to the indices of the list elements \code{"..1"}, \code{"..2"}, etc.
+#' \code{how = "unmelt"} is a special case that reconstructs a nested list from a melted data.frame. For this reason, \code{how = "unmelt"} 
+#' only applies to data.frames in the same format as returned by \code{how = "melt"}. Internally, \code{how = "unmelt"} first reconstructs 
+#' a nested list from the melted data.frame and second uses the same framework as \code{how = "replace"}.
 #' 
 #' @section Condition function:
 #' Both \code{\link{rapply}} and \code{rrapply} allow to apply \code{f} to list elements of certain classes via the \code{classes} argument. 
@@ -30,13 +33,14 @@
 #' argument need to be defined as function arguments in \emph{both} the \code{f} and \code{condition} function (if existing), even if they are not
 #'  used in the function itself. See also the \sQuote{Examples} section.
 #' 
-#' @section Special arguments \code{.xname} and \code{.xpos}:
-#' The \code{f} and \code{condition} functions accept two special arguments \code{.xname} and \code{.xpos} in addition to the first principal argument. 
-#' The \code{.xname} argument evaluates to the name of the list element. The \code{.xpos} argument evaluates to the position of the element in the nested 
-#' list structured as an integer vector. That is, if \code{x = list(list("y", "z"))}, then an \code{.xpos} location of \code{c(1, 2)} corresponds 
-#' to the list element \code{x[[c(1, 2)]]}. The names \code{.xname} and \code{.xpos} need to be explicitly included as function arguments in \code{f} and 
-#' \code{condition} (in addition to the principal argument). See the package vignette for example uses of the \code{.xname} and 
-#' \code{.xpos} variables.  
+#' @section Special arguments \code{.xname}, \code{.xpos}, \code{.xparents} and \code{.xsiblings}:
+#' The \code{f} and \code{condition} functions accept four special arguments \code{.xname}, \code{.xpos}, \code{.xparents} and \code{.xsiblings} in
+#' addition to the first principal argument. The \code{.xname} argument evaluates to the name of the list element. The \code{.xpos} argument evaluates 
+#' to the position of the element in the nested list structured as an integer vector. That is, if \code{x = list(list("y", "z"))}, then an \code{.xpos} 
+#' location of \code{c(1, 2)} corresponds to the list element \code{x[[c(1, 2)]]}. The \code{.xparents} argument evaluates to a vector of all parent 
+#' node names in the path to the list element. The \code{.xsiblings} argument evaluates to the complete (sub)list that includes the list element as a direct child.
+#' The names \code{.xname}, \code{.xpos}, \code{.xparents} or \code{.xsiblings} need to be explicitly included as function arguments in \code{f} and 
+#' \code{condition} (in addition to the principal argument). See the package vignette for example uses of these special variables.
 #' 
 #' @section List node aggregation:
 #' By default, \code{rrapply} recurses into any \dQuote{list-like} element. If \code{feverywhere = "break"}, this behavior is overridden and the 
@@ -62,15 +66,24 @@
 #' \code{how = "list"}. To avoid unexpected behavior, \code{rrapply} always preserves intermediate list attributes when using \code{how = "replace"}, 
 #' \code{how = "list"} or \code{how = "prune"}. If \code{how = "flatten"} or \code{how = "unlist"} intermediate list attributes cannot be preserved as 
 #' the result is no longer a nested list. 
-#'
+#' 
+#' @section Expressions:
+#' Call objects and expression vectors are also accepted as \code{object} argument, which are treated as nested lists based on their internal abstract
+#' syntax trees. As such, all functionality that applies to nested lists extends directly to call objects and expression vectors. If \code{object} is a 
+#' call object or expression vector, \code{how = "replace"} always maintains the type of \code{object}, whereas \code{how = "list"} returns the result 
+#' structured as a nested list. \code{how = "prune"}, \code{how = "flatten"} and \code{how = "melt"} return the pruned abstract syntax tree as: a nested list, 
+#' a flattened list and a melted data.frame respectively. This is identical to application of \code{rrapply} to the abstract syntax tree formatted as a nested list.
+#' 
 #' @return If \code{how = "unlist"}, a vector as in \code{\link{rapply}}. If \code{how = "list"} or \code{how = "replace"}, \dQuote{list-like} of similar 
 #' structure as \code{object} as in \code{\link{rapply}}. If \code{how = "prune"}, a pruned \dQuote{list-like} object of similar structure as \code{object}
 #' with pruned list elements based on \code{classes} and \code{condition}. If \code{how = "flatten"}, an unnested pruned list with pruned list elements 
 #' based on \code{classes} and \code{condition}. If \code{how = "melt"}, a melted data.frame containing the node paths and values of the pruned list 
-#' elements based on \code{classes} and \code{condition}.
+#' elements based on \code{classes} and \code{condition}. If \code{how = "unmelt"}, a nested list with list names and values defined in the data.frame \code{object}.
 #'  
 #' @note \code{rrapply} allows the \code{f} function argument to be missing, in which case no function is applied to the list 
 #' elements.
+#' @note \code{how = "unmelt"} requires as input a data.frame as returned by \code{how = "melt"} with character columns to name the nested list components
+#' and a final list-column containing the values of the nested list elements. 
 #' 
 #' @examples
 #' # Example data
@@ -80,7 +93,7 @@
 #' ## Subset values for countries and areas in Oceania
 #' renewable_oceania <- renewable_energy_by_country[["World"]]["Oceania"]
 #'
-#' # List pruning
+#' # List pruning and unnesting
 #' 
 #' ## Drop all logical NA's while preserving list structure 
 #' na_drop_oceania <- rrapply(
@@ -107,8 +120,17 @@
 #'   classes = "numeric",
 #'   how = "melt"
 #' )
+#' 
 #' head(na_drop_oceania3)
 #' 
+#' ## Reconstruct nested list from melted data.frame
+#' na_drop_oceania4 <- rrapply(
+#'   na_drop_oceania3,
+#'   how = "unmelt"
+#' )
+#' 
+#' str(na_drop_oceania4, list.len = 3, give.attr = FALSE)
+#' 
 #' # Condition function
 #' 
 #' ## Drop all NA elements using condition function
@@ -140,7 +162,7 @@
 #' )
 #' str(renewable_energy_above_85, give.attr = FALSE)
 #' 
-#' # Special arguments .xname and .xpos
+#' # Special arguments .xname, .xpos and .xparents
 #' 
 #' ## Apply a function using the name of the node
 #' renewable_oceania_text <- rrapply(
@@ -167,6 +189,14 @@
 #' )
 #' str(renewable_europe_above_50, give.attr = FALSE)
 #' 
+#' ## Filter European countries with value above 50%
+#' renewable_europe_above_50 <- rrapply(
+#'   renewable_energy_by_country,
+#'   condition = function(x, .xparents) "Europe" %in% .xparents & x > 50,
+#'   how = "prune"
+#' )
+#' str(renewable_europe_above_50, give.attr = FALSE)
+#' 
 #' ## Return position of Sweden in list
 #' (xpos_sweden <- rrapply(
 #'   renewable_energy_by_country,
@@ -176,6 +206,14 @@
 #' ))
 #' renewable_energy_by_country[[xpos_sweden$Sweden]]
 #' 
+#' ## Return siblings of Sweden in list
+#' siblings_sweden <- rrapply(
+#'   renewable_energy_by_country,
+#'   condition = function(x, .xsiblings) "Sweden" %in% names(.xsiblings),
+#'   how = "flatten"
+#' )
+#' str(siblings_sweden, list.len = 10, give.attr = FALSE)
+#' 
 #' # List node aggregation
 #' 
 #' ## Calculate mean value of Europe
@@ -231,41 +269,58 @@
 #' )
 #' str(na_drop_oceania_attr2, max.level = 2)
 #' 
-#' # Applying functions to data.frames
-#' 
-#' ## Scale only Sepal columns in iris dataset
-#' head(
-#'   rrapply(
-#'     iris,
-#'     condition = function(x, .xname) grepl("Sepal", .xname),
-#'     f = scale
-#'   )
+#' # Expressions
+#'
+#' ## Replace logicals by integers
+#' call_old <- quote(y <- x <- 1 + TRUE)
+#' call_new <- rrapply(call_old, 
+#'   classes = "logical", 
+#'   f = as.numeric, 
+#'   how = "replace"
 #' )
+#' str(call_new)
 #' 
-#' ## Scale and keep only numeric columns
-#' head(
-#'   rrapply(
-#'     iris,
-#'     f = scale,
-#'     classes = "numeric",
-#'     how = "prune"
-#'   )
+#' ## Update and decompose call object
+#' call_ast <- rrapply(call_old, 
+#'   f = function(x) ifelse(is.logical(x), as.numeric(x), x), 
+#'   how = "list"
 #' )
+#' str(call_ast)
 #' 
-#' ## Summarize only numeric columns with how = "flatten"
-#' rrapply(
-#'   iris,
-#'   f = summary,
-#'   classes = "numeric",
+#' ## Prune and decompose expression
+#' expr <- expression(y <- x <- 1, f(g(2 * pi)))
+#' is_new_name <- function(x) !exists(as.character(x), envir = baseenv())
+#' expr_prune <- rrapply(expr, 
+#'   classes = "name", 
+#'   condition = is_new_name, 
+#'   how = "prune"
+#' )
+#' str(expr_prune)
+#' 
+#' ## Prune and flatten expression
+#' expr_flatten <- rrapply(expr, 
+#'   classes = "name", 
+#'   condition = is_new_name, 
 #'   how = "flatten"
 #' )
+#' str(expr_flatten)
+#' 
+#' ## Prune and melt expression
+#' expr_melt <- rrapply(expr, 
+#'   classes = "name", 
+#'   condition = is_new_name, 
+#'   f = as.character,
+#'   how = "melt"
+#' )
+#' expr_melt
 #' 
 #' @inheritParams base::rapply
-#' @param f a \code{\link{function}} of one \dQuote{principal} argument and optional special arguments \code{.xname} and/or \code{.xpos} 
-#' (see \sQuote{Details}), passing further arguments via \code{\dots}.
-#' @param condition a condition \code{\link{function}} of one \dQuote{principal} argument and optional special arguments \code{.xname} and/or 
-#' \code{.xpos} (see \sQuote{Details}), passing further arguments via \code{\dots}.
-#' @param how character string partially matching the five possibilities given: see \sQuote{Details}.
+#' @param object a \code{\link{list}}, \code{\link{expression}} vector, or \code{\link{call}} object, i.e., \dQuote{list-like}.
+#' @param f a \code{\link{function}} of one \dQuote{principal} argument and optional special arguments \code{.xname}, \code{.xpos}, \code{.xparents} 
+#' and/or \code{.xsiblings} (see \sQuote{Details}), passing further arguments via \code{\dots}.
+#' @param condition a condition \code{\link{function}} of one \dQuote{principal} argument and optional special arguments \code{.xname}, \code{.xpos}, 
+#' \code{.xparents} and/or \code{.xsiblings} (see \sQuote{Details}), passing further arguments via \code{\dots}.
+#' @param how character string partially matching the seven possibilities given: see \sQuote{Details}.
 #' @param deflt the default result (only used if \code{how = "list"} or \code{how = "unlist"}).
 #' @param dfaslist logical value to treat data.frames as \dQuote{list-like} object.
 #' @param feverywhere character options \code{"break"} or \code{"recurse"} to override default behavior of \code{f}: see \sQuote{Details}.
@@ -279,66 +334,90 @@
 #' @useDynLib rrapply, .registration = TRUE
 #' @export 
 rrapply <- function(object, condition, f, classes = "ANY", deflt = NULL, 
-                    how = c("replace", "list", "unlist", "prune", "flatten", "melt"),
-                    feverywhere = NULL, dfaslist = TRUE, ...)
+    how = c("replace", "list", "unlist", "prune", "flatten", "melt", "unmelt"),
+    feverywhere = NULL, dfaslist = TRUE, ...)
 {
-  
   ## non-function arguments
-  if(!is.list(object) || length(object) < 1) stop("'object' argument should be list-like and of length greater than zero")
-  how <- match.arg(how, c("replace", "list", "unlist", "prune", "flatten", "melt"))
-  howInt <- match(how, c("replace", "list", "unlist", "prune", "flatten", "melt"))
+  if(!(is.list(object) || is.call(object) || is.expression(object)) || length(object) < 1) 
+    stop("'object' argument should be list-like and of length greater than zero")
+  
+  how <- match.arg(how, c("replace", "list", "unlist", "prune", "flatten", "melt", "unmelt"))
+  howInt <- match(how, c("replace", "list", "unlist", "prune", "flatten", "melt", "unmelt"))
   dfaslist <- isTRUE(dfaslist)
   feverywhere <- match.arg(feverywhere, c("no", "break", "recurse"))
   feverywhereInt <- match(feverywhere, c("no", "break", "recurse"))
   
+  ## unmelt data.frame to nested list
+  if(identical(how, "unmelt"))
+  {
+    if(is.data.frame(object) && is.list(object[[length(object)]]) && all(vapply(object[, -length(object)], is.character, logical(1))))
+    {
+      if(nrow(object) > 0 && length(object) > 1)
+      {
+        object <- .Call(C_unmelt, object)
+        how <- "replace"
+        howInt <- 1L
+      }
+      else
+        stop("'object' argument must be non-empty and contain at least one naming character column and one value list column")
+    }
+    else
+      stop("'object' argument must be a data.frame consisting of naming character columns and a value list column")
+  }
   ## function arguments  
   if(missing(f)) f <- NULL else f <- match.fun(f)
   if(missing(condition)) condition <- NULL else condition <- match.fun(condition)
   
-  if(is.null(f) && (is.null(condition) || howInt == 0L) && (feverywhereInt < 1L) && howInt < 4L) 
+  if(is.null(f) && (is.null(condition) || identical(how, "replace")) && identical(feverywhereInt, 1L) && 
+      ((is.list(object) && howInt < 5L) || (!is.list(object) && howInt < 2L)))
   {  
     ## nothing to be done
     res <- object  
   } else 
   { 
-    ## check for .xname and .xpos args
-    fArgs <- conditionArgs <- c(0L, 0L)
+    ## check for special args
+    fArgs <- conditionArgs <- c(0L, 0L, 0L, 0L)
     if(identical(typeof(f), "closure"))
-      fArgs <- match(c(".xname", ".xpos"), names(formals(f)), nomatch = 0L) 
+      fArgs <- match(c(".xname", ".xpos", ".xparents", ".xsiblings"), names(formals(f)), nomatch = 0L) 
     if(identical(typeof(condition), "closure"))
-      conditionArgs <- match(c(".xname", ".xpos"), names(formals(condition)), nomatch = 0L)
+      conditionArgs <- match(c(".xname", ".xpos", ".xparents", ".xsiblings"), names(formals(condition)), nomatch = 0L)
     
     ## call main C function
     res <- .Call(C_rrapply, environment(), object, f, fArgs, condition, conditionArgs, classes, howInt, deflt, dfaslist, feverywhereInt)  
   }
   
-  if(how == "melt")
+  if(identical(how, "melt"))
   {
     ## drop NULL name columns
     res <- res[vapply(res, Negate(is.null), logical(1L))]
     ## convert list to data.frame
     if(length(res) > 1)
     {
-      res <- structure(
-        res,
-        names =  c(paste0("L", seq_len(length(res) - 1L)), "value"),
-        row.names = seq_len(length(res[[1L]])),
-        class = "data.frame"
-      )
+      if(getRversion() >= "4.0.0" || !any(vapply(res[[length(res)]], is.symbol, logical(1L)))) {
+        res <- structure(
+            res,
+            names =  c(paste0("L", seq_len(length(res) - 1L)), "value"),
+            row.names = seq_len(length(res[[1L]])),
+            class = "data.frame"
+        )
+      } else {
+        ## no format method for names in R < 4.0.0 
+        names(res) <- c(paste0("L", seq_len(length(res) - 1L)), "value")
+      }
     }
     else
     {
       res <- structure(
-        res,
-        names = "value",
-        row.names = integer(0L),
-        class = "data.frame"
+          res,
+          names = "value",
+          row.names = integer(0L),
+          class = "data.frame"
       )
     }
   }
   
   ## unlist result
-  if(how == "unlist") res <- unlist(res, recursive = TRUE)
+  if(identical(how, "unlist")) res <- unlist(res, recursive = TRUE)
   
   return(res)
   
