@@ -15,16 +15,21 @@ typedef struct Args
 		1: list (fill default, return nested list)
 		2: unlist (fill default, return nested list, unlist in R-function)
 		3: prune (return nested list)
-		4: flatten (prune, return flat list)    
+		4: flatten (prune, return flat list)  
+		5: melt (prune, return melted list)
 	*/
 
 	int how_C;		 // how argument
 	int fArgs;		 // number of arguments f
 	int fxname;		 // .xname present in f
 	int fxpos;		 // .xpos present in f
+	int fxparents;	 // .xparents present in f
+	int fxsiblings;	 // .xsiblings present in f
 	int pArgs;		 // number of arguments condition
 	int pxname;		 // .xname present in condition
 	int pxpos;		 // .xpos present in condition
+	int pxparents;	 // .xparents present in condition
+	int pxsiblings;	 // .xsiblings present in condition
 	int dfaslist;	 // dfaslist argument
 	int feverywhere; // feverywhere argument
 } Args;
@@ -47,25 +52,24 @@ typedef struct CountLocal
 } CountLocal;
 
 /* prototypes */
-
+static SEXP C_lang7(SEXP s, SEXP t, SEXP u, SEXP v, SEXP w, SEXP x, SEXP y);
 static SEXP C_int2char(int i);
 static void C_copyAttrs(SEXP obj, SEXP ans, SEXP names, Rboolean copyAttrs);
 static int C_matchClass(SEXP obj, SEXP classes);
 static void C_traverse(SEXP X, CountGlobal *count, int depth);
-static SEXP C_eval_list(SEXP env, SEXP Xi, SEXP fcall, SEXP pcall, SEXP classes, SEXP deflt, SEXP xsym, SEXP xnameChar, Args args, CountGlobal *countglobal, CountLocal countlocal, R_len_t (**xinfo)[3], R_len_t **xloc, R_len_t **xdepth);
+static SEXP C_eval_list(SEXP env, SEXP Xi, SEXP fcall, SEXP pcall, SEXP classes, SEXP deflt, SEXP xsym, SEXP *xparents, SEXP xsiblings, PROTECT_INDEX ipx, Args args, CountGlobal *countglobal, CountLocal countlocal, R_len_t (**xinfo)[3], R_len_t **xloc, R_len_t **xdepth);
 static SEXP C_fill_list(SEXP Xi, R_len_t (*xinfo)[3], R_len_t *buf, R_len_t node, R_len_t maxnodes, R_len_t ibuf);
 static void C_fill_flat(SEXP ansNew, SEXP Xi, R_len_t (*xinfo)[3], R_len_t *ix, R_len_t *ians);
 static void C_fill_flat_names(SEXP ansNew, SEXP newNames, SEXP Xi, SEXP name, R_len_t (*xinfo)[3], R_len_t *ix, R_len_t *ians);
 static void C_fill_melt(SEXP ansFlat, SEXP ansNames, SEXP Xi, SEXP name, R_len_t (*xinfo)[3], R_len_t *ix, R_len_t *ians);
 SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPred, SEXP classes, SEXP how, SEXP deflt, SEXP R_dfaslist, SEXP R_feverywhere);
-
 /* ---------------------- */
 
 /* Main function */
 
 SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPred, SEXP classes, SEXP R_how, SEXP deflt, SEXP R_dfaslist, SEXP R_feverywhere)
 {
-	SEXP ans, names, xsym, xname, xpos, R_fcall, R_pcall;
+	SEXP ans, ansptr, xptr, names, xsym, xname, xpos, xparents, xsiblings, R_xparents, R_fcall, R_pcall;
 
 	/* protect calls */
 	int nprotect = 0;
@@ -76,9 +80,13 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 	R_args.fArgs = 0;
 	R_args.fxname = INTEGER_ELT(argsFun, 0) > 0;
 	R_args.fxpos = INTEGER_ELT(argsFun, 1) > 0;
+	R_args.fxparents = INTEGER_ELT(argsFun, 2) > 0;
+	R_args.fxsiblings = INTEGER_ELT(argsFun, 3) > 0;
 	R_args.pArgs = 0;
 	R_args.pxname = INTEGER_ELT(argsPred, 0) > 0;
 	R_args.pxpos = INTEGER_ELT(argsPred, 1) > 0;
+	R_args.pxparents = INTEGER_ELT(argsPred, 2) > 0;
+	R_args.pxsiblings = INTEGER_ELT(argsPred, 3) > 0;
 	R_args.dfaslist = LOGICAL_ELT(R_dfaslist, 0);
 	R_args.feverywhere = INTEGER_ELT(R_feverywhere, 0) - 1;
 
@@ -101,39 +109,68 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 	if (R_args.how_C == 5)
 		xdepth = (R_len_t *)S_alloc(initGlobal.maxnodes, sizeof(R_len_t));
 
+	PROTECT_INDEX parentipx;
+	if (R_args.fxparents || R_args.pxparents)
+		PROTECT_WITH_INDEX(R_xparents = Rf_allocVector(STRSXP, initGlobal.depthmax), &parentipx);
+	else
+		PROTECT_WITH_INDEX(R_xparents = Rf_ScalarString(NA_STRING), &parentipx);
+	nprotect++;
+
 	/* install arguments and initialize call objects */
 	xsym = Rf_install("X");
 	xname = Rf_install(".xname");
 	xpos = Rf_install(".xpos");
+	xparents = Rf_install(".xparents");
+	xsiblings = Rf_install(".xsiblings");
 
 	if (Rf_isFunction(FUN))
 	{
-		/* call definitions depend on presence of .xname and.xpos arguments */
-		if (R_args.fxname && R_args.fxpos)
+		/* call definitions depend on presence of special arguments */
+		R_args.fArgs = 1 + R_args.fxname + R_args.fxpos + R_args.fxparents + R_args.fxsiblings;
+
+		switch (R_args.fArgs)
 		{
-			R_args.fArgs += 3;
-			R_fcall = PROTECT(Rf_lang5(FUN, xsym, R_NilValue, R_NilValue, R_DotsSymbol));
-			SET_TAG(CDDR(R_fcall), xname);
-			SET_TAG(CDDDR(R_fcall), xpos);
-		}
-		else if (R_args.fxname)
-		{
-			R_args.fArgs += 2;
-			R_fcall = PROTECT(Rf_lang4(FUN, xsym, R_NilValue, R_DotsSymbol));
-			SET_TAG(CDDR(R_fcall), xname);
-		}
-		else if (R_args.fxpos)
-		{
-			R_args.fArgs += 2;
-			R_fcall = PROTECT(Rf_lang4(FUN, xsym, R_NilValue, R_DotsSymbol));
-			SET_TAG(CDDR(R_fcall), xpos);
-		}
-		else
-		{
-			R_args.fArgs += 1;
+		case 1:
 			R_fcall = PROTECT(Rf_lang3(FUN, xsym, R_DotsSymbol));
+			break;
+		case 2:
+			R_fcall = PROTECT(Rf_lang4(FUN, xsym, R_NilValue, R_DotsSymbol));
+			break;
+		case 3:
+			R_fcall = PROTECT(Rf_lang5(FUN, xsym, R_NilValue, R_NilValue, R_DotsSymbol));
+			break;
+		case 4:
+			R_fcall = PROTECT(Rf_lang6(FUN, xsym, R_NilValue, R_NilValue, R_NilValue, R_DotsSymbol));
+			break;
+		case 5:
+			R_fcall = PROTECT(C_lang7(FUN, xsym, R_NilValue, R_NilValue, R_NilValue, R_NilValue, R_DotsSymbol));
+			break;
+		default:
+			R_fcall = PROTECT(Rf_lang3(FUN, xsym, R_DotsSymbol)); // not reached
 		}
 		nprotect++;
+
+		SEXP fCDR = CDR(R_fcall);
+
+		if (R_args.fxname)
+		{
+			fCDR = CDR(fCDR);
+			SET_TAG(fCDR, xname);
+		}
+		if (R_args.fxpos)
+		{
+			fCDR = CDR(fCDR);
+			SET_TAG(fCDR, xpos);
+		}
+		if (R_args.fxparents)
+		{
+			fCDR = CDR(fCDR);
+			SET_TAG(CDR(fCDR), xparents);
+		}
+		if (R_args.fxsiblings)
+		{
+			SET_TAG(CDR(fCDR), xsiblings);
+		}
 	}
 	else
 	{
@@ -142,31 +179,52 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 
 	if (Rf_isFunction(PRED))
 	{
-		if (R_args.pxname && R_args.pxpos)
+		/* call definitions depend on presence of special arguments */
+		R_args.pArgs = 1 + R_args.pxname + R_args.pxpos + R_args.pxparents + R_args.pxsiblings;
+
+		switch (R_args.pArgs)
 		{
-			R_args.pArgs += 3;
-			R_pcall = PROTECT(Rf_lang5(PRED, xsym, R_NilValue, R_NilValue, R_DotsSymbol));
-			SET_TAG(CDDR(R_pcall), xname);
-			SET_TAG(CDDDR(R_pcall), xpos);
-		}
-		else if (R_args.pxname)
-		{
-			R_args.pArgs += 2;
-			R_pcall = PROTECT(Rf_lang4(PRED, xsym, R_NilValue, R_DotsSymbol));
-			SET_TAG(CDDR(R_pcall), xname);
-		}
-		else if (R_args.pxpos)
-		{
-			R_args.pArgs += 2;
-			R_pcall = PROTECT(Rf_lang4(PRED, xsym, R_NilValue, R_DotsSymbol));
-			SET_TAG(CDDR(R_pcall), xpos);
-		}
-		else
-		{
-			R_args.pArgs += 1;
+		case 1:
 			R_pcall = PROTECT(Rf_lang3(PRED, xsym, R_DotsSymbol));
+			break;
+		case 2:
+			R_pcall = PROTECT(Rf_lang4(PRED, xsym, R_NilValue, R_DotsSymbol));
+			break;
+		case 3:
+			R_pcall = PROTECT(Rf_lang5(PRED, xsym, R_NilValue, R_NilValue, R_DotsSymbol));
+			break;
+		case 4:
+			R_pcall = PROTECT(Rf_lang6(PRED, xsym, R_NilValue, R_NilValue, R_NilValue, R_DotsSymbol));
+			break;
+		case 5:
+			R_pcall = PROTECT(C_lang7(PRED, xsym, R_NilValue, R_NilValue, R_NilValue, R_NilValue, R_DotsSymbol));
+			break;
+		default:
+			R_pcall = PROTECT(Rf_lang3(PRED, xsym, R_DotsSymbol)); // not reached
 		}
 		nprotect++;
+
+		SEXP pCDR = CDR(R_pcall);
+
+		if (R_args.pxname)
+		{
+			pCDR = CDR(pCDR);
+			SET_TAG(pCDR, xname);
+		}
+		if (R_args.pxpos)
+		{
+			pCDR = CDR(pCDR);
+			SET_TAG(pCDR, xpos);
+		}
+		if (R_args.pxparents)
+		{
+			pCDR = CDR(pCDR);
+			SET_TAG(CDR(pCDR), xparents);
+		}
+		if (R_args.pxsiblings)
+		{
+			SET_TAG(CDR(pCDR), xsiblings);
+		}
 	}
 	else
 	{
@@ -180,14 +238,23 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 		initGlobal.anynames = TRUE;
 
 	/* allocate output list */
-	if (R_args.how_C == 0 || R_args.how_C > 2)
+	ansptr = NULL; /* avoid unitialized warning */
+	xptr = NULL;
+	if (Rf_isPairList(X))
+		xptr = X;
+
+	if ((Rf_isVectorList(X) && (R_args.how_C == 1 || R_args.how_C == 2)) || (Rf_isPairList(X) && R_args.how_C > 0))
 	{
-		ans = PROTECT(Rf_shallow_duplicate(X));
+		ans = PROTECT(Rf_allocVector(VECSXP, n));
+		C_copyAttrs(X, ans, names, !Rf_isPairList(X));
+		if (Rf_isPairList(X))
+			Rf_copyMostAttrib(X, ans);
 	}
 	else
 	{
-		ans = PROTECT(Rf_allocVector((SEXPTYPE)TYPEOF(X), n));
-		C_copyAttrs(X, ans, names, TRUE);
+		ans = PROTECT(Rf_shallow_duplicate(X));
+		if (Rf_isPairList(X))
+			ansptr = ans;
 	}
 	nprotect += 2;
 
@@ -196,6 +263,8 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 	{
 		/* increment location counter */
 		xloc[0] += 1;
+		/* update parent names */
+		SET_STRING_ELT(R_xparents, 0, Rf_isNull(names) ? NA_STRING : STRING_ELT(names, i));
 
 		/* update current node info for list pruning */
 		if (R_args.how_C > 2)
@@ -218,7 +287,23 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 		}
 
 		/* main recursion part */
-		SET_VECTOR_ELT(ans, i, C_eval_list(env, VECTOR_ELT(X, i), R_fcall, R_pcall, classes, deflt, xsym, Rf_isNull(names) ? NA_STRING : STRING_ELT(names, i), R_args, &initGlobal, initLocal, &xinfo, &xloc, &xdepth));
+		if (Rf_isVectorList(X))
+		{
+			SET_VECTOR_ELT(ans, i, C_eval_list(env, VECTOR_ELT(X, i), R_fcall, R_pcall, classes, deflt, xsym, &R_xparents, (R_args.fxsiblings || R_args.pxsiblings) ? X : R_NilValue, parentipx, R_args, &initGlobal, initLocal, &xinfo, &xloc, &xdepth));
+		}
+		else if (Rf_isPairList(X))
+		{
+			if (R_args.how_C > 0)
+			{
+				SET_VECTOR_ELT(ans, i, C_eval_list(env, CAR(xptr), R_fcall, R_pcall, classes, deflt, xsym, &R_xparents, (R_args.fxsiblings || R_args.pxsiblings) ? X : R_NilValue, parentipx, R_args, &initGlobal, initLocal, &xinfo, &xloc, &xdepth));
+			}
+			else
+			{
+				SETCAR(ansptr, C_eval_list(env, CAR(xptr), R_fcall, R_pcall, classes, deflt, xsym, &R_xparents, (R_args.fxsiblings || R_args.pxsiblings) ? X : R_NilValue, parentipx, R_args, &initGlobal, initLocal, &xinfo, &xloc, &xdepth));
+				ansptr = CDR(ansptr);
+			}
+			xptr = CDR(xptr);
+		}
 	}
 
 	/* list pruning */
@@ -232,7 +317,7 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 		{
 			/* if nested list filter only level zero evaluated nodes, 
 			   otherwise filter evaluated terminal nodes */
-			if (R_args.how_C == 3 ? (xinfo[i][0] && xinfo[i][1] == -1) : (xinfo[i][0] == 1)) 
+			if (R_args.how_C == 3 ? (xinfo[i][0] && xinfo[i][1] == -1) : (xinfo[i][0] == 1))
 			{
 				buf[m] = i;
 				m++;
@@ -303,7 +388,7 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 			SEXP ansNames = PROTECT(Rf_allocVector(STRSXP, initGlobal.maxnodes));
 			PROTECT_WITH_INDEX(namesNew = Rf_getAttrib(ans, R_NamesSymbol), &ipx);
 			nprotect += 4;
-			
+
 			/* extract all evaluated parent names + populate flat list */
 			Rboolean noNames = Rf_isNull(namesNew);
 			if (noNames)
@@ -345,10 +430,9 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 					}
 				}
 				// deep copy of column
-				if(keep)
+				if (keep)
 					SET_VECTOR_ELT(ansNew, depth, Rf_duplicate(ansColumn));
 			}
-
 		}
 
 		UNPROTECT(nprotect);
@@ -363,10 +447,19 @@ SEXP C_rrapply(SEXP env, SEXP X, SEXP FUN, SEXP argsFun, SEXP PRED, SEXP argsPre
 
 /* Helper functions */
 
+/* create long language object */
+static SEXP C_lang7(SEXP s, SEXP t, SEXP u, SEXP v, SEXP w, SEXP x, SEXP y)
+{
+	PROTECT(s);
+	s = Rf_lcons(s, Rf_list6(t, u, v, w, x, y));
+	UNPROTECT(1);
+	return s;
+}
+
 /* convert integer to character */
 static SEXP C_int2char(int i)
 {
-	char buf[100];  // fixed buffer size
+	char buf[100]; // fixed buffer size
 	snprintf(buf, 100, "..%d", i);
 	return Rf_mkChar(buf);
 }
@@ -456,16 +549,26 @@ static int C_matchClass(SEXP obj, SEXP classes)
 static void C_traverse(SEXP X, CountGlobal *count, int depth)
 {
 	SEXP Xi;
+	SEXP xptr = X;
 	/* increment max depth if current depth is higher than max depth */
 	R_len_t n = Rf_length(X);
 	depth++;
 	count->maxnodes += n;
 	count->depthmax += (depth > count->depthmax);
+
 	for (R_len_t i = 0; i < n; i++)
 	{
-		Xi = VECTOR_ELT(X, i);
+		if (Rf_isVectorList(X))
+		{
+			Xi = VECTOR_ELT(X, i);
+		}
+		else
+		{
+			Xi = CAR(xptr);
+			xptr = CDR(xptr);
+		}
 		/* descend one level */
-		if (Rf_isVectorList(Xi))
+		if (TYPEOF(Xi) != NILSXP && (Rf_isVectorList(Xi) || Rf_isPairList(Xi))) // skip NILSXP
 		{
 			C_traverse(Xi, count, depth);
 		}
@@ -480,7 +583,9 @@ static SEXP C_eval_list(
 	SEXP classes,			  // classes argument
 	SEXP deflt,				  // deflt argument
 	SEXP xsym,				  // principal argument symbol
-	SEXP xnameChar,			  // current value .xname argument
+	SEXP *xparents,			  // current value .xname and .xparents arguments
+	SEXP xsiblings,			  // current value .xsiblings argument
+	PROTECT_INDEX ipx,		  // protection index xparents
 	Args args,				  // integer arguments
 	CountGlobal *countglobal, // global node counters
 	CountLocal countlocal,	  // local node counters
@@ -489,14 +594,17 @@ static SEXP C_eval_list(
 	R_len_t **xdepth		  // current depth (only used for melting)
 )
 {
-	SEXP funVal = NULL; /* avoid unitialized warning */
-	int nprotect = 0;
+	/* initialize function value */
+	SEXP funVal;
+	PROTECT_INDEX childipx;
+	PROTECT_WITH_INDEX(funVal = Rf_lazy_duplicate(Xi), &childipx);
+	int nprotect = 1;
 
 	/* if Xi is list (and data.frame is treated as list if !dfaslist)
 	   and !feverywhere recurse, otherwise evaluate functions */
 	int doRecurse = 0;
 
-	if (args.feverywhere < 1 && Rf_isVectorList(Xi))
+	if (args.feverywhere < 1 && ((Rf_isVectorList(Xi) || Rf_isPairList(Xi)) && TYPEOF(Xi) != NILSXP))
 	{
 		doRecurse = 1;
 		if (!args.dfaslist)
@@ -510,57 +618,142 @@ static SEXP C_eval_list(
 
 	if (!doRecurse)
 	{
-		SEXP xname_val, xpos_val;
+		SEXP xname_val = NULL; /* avoid uninitialized warnings */
+		SEXP xpos_val = NULL;
+		SEXP xparents_val = NULL;
+		SEXP xsiblings_val = NULL;
+		int nargprotect = 0;
 
 		/* define X argument */
 		Rf_defineVar(xsym, Xi, env);
 		INCREMENT_NAMED(Xi);
 
-		if (args.fxname > 0 || args.pxname > 0)
+		/* update current .xname value */
+		if (args.fxname || args.pxname)
 		{
-			/* update current .xname value */
-			xname_val = PROTECT(Rf_ScalarString(xnameChar));
-			if (args.fArgs > 1 && args.fxname > 0)
-				SETCADDR(fcall, xname_val);
-			if (args.pArgs > 1 && args.pxname > 0)
-				SETCADDR(pcall, xname_val);
-			UNPROTECT(1);
+			if (args.fxparents || args.pxparents)
+				xname_val = PROTECT(Rf_ScalarString(STRING_ELT(*xparents, countlocal.depth)));
+			else
+				xname_val = PROTECT(Rf_duplicate(*xparents));
+			nargprotect++;
 		}
 
-		if (args.fxpos > 0 || args.pxpos > 0)
+		/* update current .xpos value */
+		if (args.fxpos || args.pxpos)
 		{
-			/* update current .xpos value */
 			xpos_val = PROTECT(Rf_allocVector(INTSXP, countlocal.depth + 1));
 			for (R_len_t k = 0; k < (countlocal.depth + 1); k++)
 				SET_INTEGER_ELT(xpos_val, k, (int)((*xloc)[k]));
-
-			if (args.fArgs > 1 && args.fxpos > 0)
-			{
-				if (args.fxname > 0)
-					SETCADDDR(fcall, xpos_val);
-				else
-					SETCADDR(fcall, xpos_val);
-			}
-			if (args.pArgs > 1 && args.pxpos > 0)
-			{
-				if (args.pxname > 0)
-					SETCADDDR(pcall, xpos_val);
-				else
-					SETCADDR(pcall, xpos_val);
-			}
-			UNPROTECT(1);
+			nargprotect++;
 		}
 
+		/* update current .xparents value */
+		if (args.fxparents || args.pxparents)
+		{
+			xparents_val = PROTECT(Rf_allocVector(STRSXP, countlocal.depth + 1));
+			for (R_len_t k = 0; k < (countlocal.depth + 1); k++)
+				SET_STRING_ELT(xparents_val, k, STRING_ELT(*xparents, k));
+			nargprotect++;
+		}
+
+		/* update current .xsiblings value */
+		if (args.fxsiblings || args.pxsiblings)
+		{
+			if(Rf_isPairList(xsiblings))
+			{
+				R_len_t n = Rf_length(xsiblings);
+				xsiblings_val = PROTECT(Rf_allocVector(VECSXP, n));
+				SEXP xsiblingsptr = xsiblings;
+				for (R_len_t i = 0; i < n; i++)
+				{
+					SET_VECTOR_ELT(xsiblings_val, i, CAR(xsiblingsptr));
+					xsiblingsptr = CDR(xsiblingsptr);
+				}
+				Rf_copyMostAttrib(xsiblings, xsiblings_val);
+				Rf_setAttrib(xsiblings_val, R_NamesSymbol, PROTECT(Rf_getAttrib(xsiblings, R_NamesSymbol)));
+				UNPROTECT(1);
+			}
+			else 
+			{
+				xsiblings_val = PROTECT(Rf_duplicate(xsiblings));
+			}
+			nargprotect++;
+		}
+
+		/* define f special arguments */
+		if (args.fArgs > 1)
+		{
+			SEXP fCDR = CDDR(fcall);
+
+			if (args.fxname)
+			{
+				SETCAR(fCDR, xname_val);
+				fCDR = CDR(fCDR);
+			}
+			if (args.fxpos)
+			{
+				SETCAR(fCDR, xpos_val);
+				fCDR = CDR(fCDR);
+			}
+			if (args.fxparents)
+			{
+				SETCAR(fCDR, xparents_val);
+				fCDR = CDR(fCDR);
+			}
+			if (args.fxsiblings)
+			{
+				SETCAR(fCDR, xsiblings_val);
+			}
+		}
+
+		/* define condition special arguments */
+		if (args.pArgs > 1)
+		{
+			SEXP pCDR = CDDR(pcall);
+
+			if (args.pxname)
+			{
+				SETCAR(pCDR, xname_val);
+				pCDR = CDR(pCDR);
+			}
+			if (args.pxpos)
+			{
+				SETCAR(pCDR, xpos_val);
+				pCDR = CDR(pCDR);
+			}
+			if (args.pxparents)
+			{
+				SETCAR(pCDR, xparents_val);
+				pCDR = CDR(pCDR);
+			}
+			if (args.pxsiblings)
+			{
+				SETCAR(pCDR, xsiblings_val);
+			}
+		}
+
+		UNPROTECT(nargprotect);
+
 		/* evaluate predicate */
 		int doEval = TRUE;
 		int matched = FALSE;
+		int skip = FALSE;
 
+		/* skip empty symbols */
+		if (Rf_isSymbol(Xi))
+		{
+			const char *str = CHAR(PRINTNAME(Xi));
+			if (strlen(str) < 1)
+				skip = TRUE;
+		}
+
+		/* match classes argument */
 		if (strcmp(CHAR(STRING_ELT(classes, 0)), "ANY") == 0) /* ASCII */
 			matched = TRUE;
 		else
 			matched = C_matchClass(Xi, classes);
 
-		if (args.pArgs > 0)
+		if (args.pArgs > 0 && !skip)
 		{
 			/* set default to FALSE */
 			doEval = FALSE;
@@ -580,7 +773,7 @@ static SEXP C_eval_list(
 		}
 
 		/* evaluate f and decide what to return or recurse further */
-		if (doEval && matched)
+		if (doEval && matched && !skip)
 		{
 			/* update current node info only for pruning and melting */
 			if (args.how_C > 2)
@@ -594,7 +787,7 @@ static SEXP C_eval_list(
 					if (i2 > -1)
 					{
 						i1 = i2;
-						(*xinfo)[i1][0] = 2; 
+						(*xinfo)[i1][0] = 2;
 						i2 = (*xinfo)[i1][1];
 					}
 					else
@@ -605,19 +798,10 @@ static SEXP C_eval_list(
 			/* evaluate f */
 			if (args.fArgs > 0)
 			{
-				funVal = PROTECT(R_forceAndCall(fcall, args.fArgs, env));
-
-				if (MAYBE_REFERENCED(funVal))
-					funVal = Rf_lazy_duplicate(funVal);
+				REPROTECT(funVal = R_forceAndCall(fcall, args.fArgs, env), childipx);
 			}
-			else
-			{
-				funVal = PROTECT(Rf_lazy_duplicate(Xi));
-			}
-			nprotect++;
-
 			/* recurse further with new list (type 2) if feverywhere == 2 */
-			if (args.feverywhere == 2 && Rf_isVectorList(funVal))
+			if (args.feverywhere == 2 && ((Rf_isVectorList(funVal) || Rf_isPairList(funVal)) && TYPEOF(funVal) != NILSXP))
 			{
 				doRecurse = 2;
 			}
@@ -627,7 +811,7 @@ static SEXP C_eval_list(
 				return funVal;
 			}
 		}
-		else if (args.feverywhere > 0 && Rf_isVectorList(Xi))
+		else if (args.feverywhere > 0 && !skip && ((Rf_isVectorList(Xi) || Rf_isPairList(Xi)) && TYPEOF(Xi) != NILSXP))
 		{
 			/* recurse further with original list (type 1) */
 			doRecurse = 1;
@@ -635,44 +819,43 @@ static SEXP C_eval_list(
 		else
 		{
 			/* return original list (or default) here if feverywhere == 0 */
-			if (args.how_C == 0 || args.how_C > 2)
-			{
-				return Rf_lazy_duplicate(Xi);
-			}
-			else /* fill list by default */
+			if (args.how_C == 1 || args.how_C == 2)
 			{
+				UNPROTECT(nprotect);
 				return Rf_lazy_duplicate(deflt);
 			}
+			else
+			{
+				UNPROTECT(nprotect);
+				return funVal;
+			}
 		}
 	}
 
 	if (doRecurse > 0)
 	{
-		SEXP Xnew, names;
-		R_len_t m;
+		/* create new object for recursion */
+		SEXP Xnew;
+		SEXP xnewptr = NULL; /* avoid unitialized warning */
+		SEXP funptr = NULL;
+		if (Rf_isPairList(funVal))
+			funptr = funVal;
 
-		/* create new object for recursion only if doRecurse != 2 */
-		if (doRecurse != 2)
+		R_len_t m = Rf_length(funVal);
+		SEXP names = PROTECT(Rf_getAttrib(funVal, R_NamesSymbol));
+		if ((Rf_isVectorList(funVal) && (args.how_C == 1 || args.how_C == 2)) || (Rf_isPairList(funVal) && args.how_C > 0))
 		{
-			m = Rf_length(Xi);
-			names = PROTECT(Rf_getAttrib(Xi, R_NamesSymbol));
-
-			if (args.how_C == 0 || args.how_C > 2)
-			{
-				Xnew = PROTECT(Rf_shallow_duplicate(Xi));
-			}
-			else
-			{
-				/* VECEXP initializes with R_NilValues */
-				Xnew = PROTECT(Rf_allocVector(VECSXP, m));
-				C_copyAttrs(Xi, Xnew, names, TRUE);
-			}
+			/* VECEXP initializes with R_NilValues */
+			Xnew = PROTECT(Rf_allocVector(VECSXP, m));
+			C_copyAttrs(funVal, Xnew, names, !Rf_isPairList(funVal));
+			if (Rf_isPairList(funVal))
+				Rf_copyMostAttrib(funVal, Xnew);
 		}
 		else
 		{
-			m = Rf_length(funVal);
-			names = PROTECT(Rf_getAttrib(funVal, R_NamesSymbol));
 			Xnew = PROTECT(Rf_shallow_duplicate(funVal));
+			if (Rf_isPairList(funVal))
+				xnewptr = Xnew;
 		}
 		nprotect += 2;
 
@@ -689,7 +872,7 @@ static SEXP C_eval_list(
 		/* descend one level */
 		countlocal.depth++;
 
-		if (args.feverywhere == 2) 
+		if (args.feverywhere == 2)
 		{
 			if (countlocal.depth > 100) /* stop with error if depth too large */
 			{
@@ -697,7 +880,18 @@ static SEXP C_eval_list(
 			}
 			if (countlocal.depth >= countglobal->depthmax)
 			{
+				/* reallocate location vectors */
 				*xloc = (R_len_t *)S_realloc((char *)*xloc, 2 * countglobal->depthmax, countglobal->depthmax, sizeof(R_len_t));
+
+				if (args.fxparents || args.pxparents)
+				{
+					SEXP xparentsNew = PROTECT(Rf_allocVector(STRSXP, 2 * countglobal->depthmax));
+					for (R_len_t i = 0; i < countglobal->depthmax; i++)
+						SET_STRING_ELT(xparentsNew, i, STRING_ELT(*xparents, i));
+
+					REPROTECT(*xparents = xparentsNew, ipx);
+					UNPROTECT(1);
+				}
 				countglobal->depthmax *= 2;
 			}
 		}
@@ -732,15 +926,26 @@ static SEXP C_eval_list(
 
 			/* increment location */
 			(*xloc)[countlocal.depth] = j + 1;
+			/* update parent names */
+			SET_STRING_ELT(*xparents, (args.fxparents || args.pxparents) ? countlocal.depth : 0, Rf_isNull(names) ? NA_STRING : STRING_ELT(names, j));
 
 			/* evaluate list element */
-			if (doRecurse != 2)
+			if (Rf_isVectorList(funVal))
 			{
-				SET_VECTOR_ELT(Xnew, j, C_eval_list(env, VECTOR_ELT(Xi, j), fcall, pcall, classes, deflt, xsym, Rf_isNull(names) ? NA_STRING : STRING_ELT(names, j), args, countglobal, countlocal, xinfo, xloc, xdepth));
+				SET_VECTOR_ELT(Xnew, j, C_eval_list(env, VECTOR_ELT(funVal, j), fcall, pcall, classes, deflt, xsym, xparents, (args.fxsiblings || args.pxsiblings) ? funVal : R_NilValue, ipx, args, countglobal, countlocal, xinfo, xloc, xdepth));
 			}
-			else
+			else if (Rf_isPairList(funVal))
 			{
-				SET_VECTOR_ELT(Xnew, j, C_eval_list(env, VECTOR_ELT(funVal, j), fcall, pcall, classes, deflt, xsym, Rf_isNull(names) ? NA_STRING : STRING_ELT(names, j), args, countglobal, countlocal, xinfo, xloc, xdepth));
+				if (args.how_C > 0)
+				{
+					SET_VECTOR_ELT(Xnew, j, C_eval_list(env, CAR(funptr), fcall, pcall, classes, deflt, xsym, xparents, (args.fxsiblings || args.pxsiblings) ? funVal : R_NilValue, ipx, args, countglobal, countlocal, xinfo, xloc, xdepth));
+				}
+				else
+				{
+					SETCAR(xnewptr, C_eval_list(env, CAR(funptr), fcall, pcall, classes, deflt, xsym, xparents, (args.fxsiblings || args.pxsiblings) ? funVal : R_NilValue, ipx, args, countglobal, countlocal, xinfo, xloc, xdepth));
+					xnewptr = CDR(xnewptr);
+				}
+				funptr = CDR(funptr);
 			}
 		}
 
@@ -861,7 +1066,7 @@ static void C_fill_melt(SEXP ansFlat, SEXP ansNames, SEXP Xi, SEXP name, R_len_t
 {
 	// add name to vector
 	SET_STRING_ELT(ansNames, *ix, name);
-	
+
 	if (xinfo[*ix][0] == 1) // terminal nodes only
 	{
 		SET_VECTOR_ELT(ansFlat, *ians, Xi);
@@ -875,16 +1080,16 @@ static void C_fill_melt(SEXP ansFlat, SEXP ansNames, SEXP Xi, SEXP name, R_len_t
 		PROTECT_WITH_INDEX(names = Rf_getAttrib(Xi, R_NamesSymbol), &ipx);
 		Rboolean noNames = Rf_isNull(names);
 
-		if(noNames)
+		if (noNames)
 			REPROTECT(names = Rf_allocVector(STRSXP, m), ipx);
 
 		for (R_len_t i = 0; i < m; i++)
 		{
 			(*ix)++;
 			// use counter for missing names
-			if(noNames)
+			if (noNames)
 				SET_STRING_ELT(names, i, C_int2char(i + 1));
-			
+
 			// recurse further
 			C_fill_melt(ansFlat, ansNames, VECTOR_ELT(Xi, i), STRING_ELT(names, i), xinfo, ix, ians);
 		}
